package ch.lantersoft.ta;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.ArrayList;
import java.util.Random;

import javax.swing.JApplet;
import javax.swing.JComponent;
import javax.swing.JFrame;
import javax.swing.JPanel;
import javax.swing.JSpinner;
import javax.swing.SpinnerNumberModel;
import javax.swing.SwingUtilities;
import javax.swing.Timer;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

/**
 * Copyright 2010 Martin Lanter, Lantersoft. All rights reserved.
 */
public class TimerApplet extends JApplet {

	// setup frame
	public static void main(String[] args) {
		SwingUtilities.invokeLater(new Runnable() {
			public void run() {
				TimerPanel tp = new TimerPanel();
				TimerControll tc = new TimerControll(tp);
				JFrame frame = new JFrame("Performance Test Frame");
				frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
				frame.add(tp,BorderLayout.CENTER);
				frame.add(tc,BorderLayout.SOUTH);
				frame.pack();
				frame.setLocationRelativeTo(null);
				frame.setVisible(true);
			}
		});
	}

	// setup applet
	public void init() {
		SwingUtilities.invokeLater(new Runnable() {
			public void run() {
				TimerPanel tp = new TimerPanel();
				TimerControll tc = new TimerControll(tp);
				getContentPane().add(tp,BorderLayout.CENTER);
				getContentPane().add(tc,BorderLayout.SOUTH);
				setSize(getPreferredSize());
			}
		});
	}

	/**
	 * JSpinner to controll the amount of balls in TimerPanel
	 */
	private static class TimerControll extends JSpinner {

		private TimerControll(final TimerPanel tp) {
			super(new SpinnerNumberModel(tp.getBallCount(),0,10000,1));
			addChangeListener(new ChangeListener() {
				public void stateChanged(ChangeEvent e) {
					int val = (Integer)getValue();
					int n = tp.getBallCount();
					if (val < n) {
						tp.removeBalls(n-val);
					} else {
						tp.addBalls(val-n);
					}
				}
			});
		}
	}

	/**
	 * The main panel for the balls, where also the time is measured
	 */
	private static class TimerPanel extends JPanel {

		private static final int MAX_SPEED = 250; // pixels per second
		private static final int INIT_BALLS = 1000;
		private static final int INIT_DELAY = 20; // ms

		// step time
		private long locdt; // measured in ns

		private ArrayList<Ball> balls;

		public TimerPanel() {
			// setup panel
			setLayout(null);
			setBackground(Color.WHITE);
			setSize(600,400);
			setPreferredSize(new Dimension(600,400));
			setFont(getFont().deriveFont(Font.BOLD));

			// init balls
			balls = new ArrayList<Ball>();
			addBalls(INIT_BALLS);

			// init timer
			new Timer(INIT_DELAY,new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					step(INIT_DELAY);
					repaint();
				}
			}).start();
		}

		@Override
		public void paint(Graphics g) {
			long t0 = System.nanoTime();
			super.paint(g);
			long dt = System.nanoTime()-t0;
			g.setColor(Color.RED);
			g.drawString("Blle: "+getBallCount(),0,12);
			g.drawString("setBounds: "+locdt/1000000+" ms",0,24);
			g.drawString("paint: "+dt/1000000+" ms",0,36);
		}

		/**
		 * Move all balls as far as they come in the time difference dt. Sum up
		 * the needed time for all the setBounds calls
		 * @param dt - time difference, how far the balls should move
		 */
		private void step(long dt) {
			locdt = 0; // reset time
			for (Ball ball:balls) {

				locdt += ball.step(dt); // sum time for setBounds calls

				// collision with border handling
				if (ball.vx<0 && ball.getX()<0) ball.vx = -ball.vx;
				if (ball.vy<0 && ball.getY()<0) ball.vy = -ball.vy;
				if (ball.vx>0 && ball.getX()+ball.size>=getWidth()) ball.vx = -ball.vx;
				if (ball.vy>0 && ball.getY()+ball.size>=getHeight()) ball.vy = -ball.vy;
			}
		}

		/**
		 * Add m balls to the panel
		 */
		public void addBalls(int m) {
			Random rand = new Random();
			for (int i=0;i<m;i++) {
				Ball ball = new Ball(rand.nextInt(getWidth()),
						rand.nextInt(getHeight()),
						rand.nextInt(MAX_SPEED*2)-MAX_SPEED,
						rand.nextInt(MAX_SPEED*2)-MAX_SPEED,
						rand.nextInt(20)+4,
						new Color(rand.nextInt(256*256*256)));
				balls.add(ball);
				add(ball);
			}
		}

		/**
		 * Remove m balls from the panel
		 */
		public void removeBalls(int m) {
			int n = getBallCount()-1;
			for (int i=0;i<m;i++) {
				Ball b = balls.remove(n-i);
				remove(b);
			}
			n += m;
		}

		public int getBallCount() {
			return balls.size();
		}
	}

	/**
	 * Class of the balls
	 */
	private static class Ball extends JComponent {

		private float x, y;
		private float vx, vy;
		private int size;
		private Color col;

		public Ball(int x, int y, float vx, float vy, int size, Color col) {
			this.x = x;
			this.y = y;
			this.vx = vx;
			this.vy = vy;
			this.size = size;
			this.col = col;
			setLocation(x, y);
			setSize(size, size);
		}

		/**
		 * Move as far as the ball comes in the time difference dt and return
		 * the time needed for the setBounds call.
		 * @param dt - time difference, how far the balls should move
		 * @return - needed time for setBounds call
		 */
		private long step(long dt) {
			x = x+vx*dt/1000;
			y = y+vy*dt/1000;

			int xx = (int) Math.round(x);
			int yy = (int) Math.round(y);

			// meassure time for setBounds
			long t0 = System.nanoTime();
			setBounds(xx, yy, size, size);
			return System.nanoTime()-t0; // time for setBounds
		}

		public void paintComponent(Graphics g) {
			g.setColor(col);
			g.fillOval(0,0,size,size);
		}
	}
}
